/*******************************************************************************
 Infomap software package for multi-level network clustering
 Copyright (c) 2013, 2014 Daniel Edler, Anton Holmgren, Martin Rosvall

 This file is part of the Infomap software package.
 See file LICENSE_AGPLv3.txt for full license details.
 For more information, see <http://www.mapequation.org>
 ******************************************************************************/

#ifndef VERSION_H_
#define VERSION_H_

namespace infomap {

const char* const INFOMAP_VERSION = "2.7.0";

}

#endif // VERSION_H_
